# -*- coding: utf-8 -*-
"""Houdini Engine's internal data types
Author  : Maajor
Email   : info@ma-yidong.com
"""
from ctypes import Structure, Array,\
    c_int, c_int32, c_int64, c_float, c_bool
import enum
import numpy as np

class StructureWithEnums(Structure):
    """Add missing enum feature to ctypes Structures.
    ref:  https://gist.github.com/christoph2/9c390e5c094796903097
    """
    _map = {}

    def __getattribute__(self, name):
        _map = Structure.__getattribute__(self, '_map')
        value = Structure.__getattribute__(self, name)
        if name in _map:
            enum_class = _map[name]
            if isinstance(value, Array):
                return [enum_class(x) for x in value]
            return enum_class(value)
        return value

    def __str__(self):
        result = []
        result.append("struct {0} {{".format(self.__class__.__name__))
        for field in self._fields_:
            attr, attr_type = field
            if attr in self._map:
                attr_type = self._map[attr]
            value = getattr(self, attr)
            result.append("    {0} [{1}] = {2!r}".format(
                attr, attr_type.__name__, value))
        result.append("}")
        return '\n'.join(result)

    __repr__ = __str__


class LicenseType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_License
    """
    NONE = 0
    HOUDINI_ENGINE = 1
    HOUDINI = 2
    HOUDINI_FX = 3
    HOUDINI_ENGINE_INDIE = 4
    HOUDINI_INDIE = 5
    MAX = 6


class StatusType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_StatusType
    """
    CALL_RESULT = 0
    COOK_RESULT = 1
    COOK_STATE = 2
    MAX = 3


class StatusVerbosity(enum.IntEnum):
    """Equivalent of HAPI's HAPI_StatusVerbosity
    """
    ALL = 2
    ERRORS = 0
    WARNINGS = 1
    MESSAGES = 2


class Result(enum.IntEnum):
    """Equivalent of HAPI's HAPI_Result
    """
    SUCCESS = 0
    FAILURE = 1
    ALREADY_INITIALIZED = 2
    NOT_INITIALIZED = 3
    CANT_LOADFILE = 4
    PARM_SET_FAILED = 5
    INVALID_ARGUMENT = 6
    CANT_LOAD_GEO = 7
    CANT_GENERATE_PRESET = 8
    CANT_LOAD_PRESET = 9
    ASSET_DEF_ALREADY_LOADED = 10

    NO_LICENSE_FOUND = 110
    DISALLOWED_NC_LICENSE_FOUND = 120
    DISALLOWED_NC_ASSET_WITH_C_LICENSE = 130
    DISALLOWED_NC_ASSET_WITH_LC_LICENSE = 140
    DISALLOWED_LC_ASSET_WITH_C_LICENSE = 150
    DISALLOWED_HENGINEINDIE_W_3PARTY_PLUGIN = 160

    ASSET_INVALID = 200
    NODE_INVALID = 210

    USER_INTERRUPTED = 300

    INVALID_SESSION = 400


class ErrorCode(enum.IntEnum):
    """Equivalent of HAPI's HAPI_ErrorCode
    """
    ASSET_DEF_NOT_FOUND = 1 << 0
    PYTHON_NODE_ERROR = 1 << 1


class SessionType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_SessionType
    """
    INPROCESS = 0
    THRIFT = 1
    CUSTOM1 = 2
    CUSTOM2 = 3
    CUSTOM3 = 4
    MAX = 5


class State(enum.IntEnum):
    """Equivalent of HAPI's HAPI_State
    """
    READY = 0
    READY_WITH_FATAL_ERRORS = 1
    READY_WITH_COOK_ERRORS = 2
    STARTING_COOK = 3
    COOKING = 4
    STARTING_LOAD = 5
    LOADING = 6
    MAX = 7

    MAX_READY_STATE = READY_WITH_COOK_ERRORS


class PackedPrimInstancingMode(enum.IntEnum):
    """Equivalent of HAPI's HAPI_PackedPrimInstancingMode
    """
    INVALID = -1
    DISABLED = 0
    HIERARCHY = 1
    FLAT = 2
    MAX = 3


class Permissions(enum.IntEnum):
    """Equivalent of HAPI's HAPI_Permissions
    """
    NON_APPLICABLE = 0
    READ_WRITE = 1
    READ_ONLY = 2
    WRITE_ONLY = 3
    MAX = 4


class RampType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_RampType
    """
    INVALID = -1
    FLOAT = 0
    COLOR = 1
    MAX = 2


class ParmType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_ParmType
    """
    INT = 0
    MULTIPARMLIST = 1
    TOGGLE = 2
    BUTTON = 3

    FLOAT = 4
    COLOR = 5

    STRING = 6
    PATH_FILE = 7
    PATH_FILE_GEO = 8
    PATH_FILE_IMAGE = 9

    NODE = 10

    FOLDERLIST = 11
    FOLDERLIST_RADIO = 12

    FOLDER = 13
    LABEL = 14
    SEPARATOR = 15
    PATH_FILE_DIR = 16

    MAX = 17

    INT_START = INT
    INT_END = BUTTON

    FLOAT_START = FLOAT
    FLOAT_END = COLOR

    STRING_START = STRING
    STRING_END = NODE

    PATH_START = PATH_FILE
    PATH_END = PATH_FILE_IMAGE

    NODE_START = NODE
    NODE_END = NODE

    CONTAINER_START = FOLDERLIST
    CONTAINER_END = FOLDERLIST_RADIO

    NONVALUE_START = FOLDER
    NONVALUE_END = SEPARATOR


class PrmScriptType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_PrmScriptType
    """
    INT = 0
    FLOAT = 1
    ANGLE = 2
    STRING = 3
    FILE = 4
    DIRECTORY = 5
    IMAGE = 6
    GEOMETRY = 7
    TOGGLE = 8
    BUTTON = 9
    VECTOR2 = 10
    VECTOR3 = 11
    VECTOR4 = 12
    INTVECTOR2 = 13
    INTVECTOR3 = 14
    INTVECTOR4 = 15
    UV = 16
    UVW = 17
    DIR = 18
    COLOR = 19
    COLOR4 = 20
    OPPATH = 21
    OPLIST = 22
    OBJECT = 23
    OBJECTLIST = 24
    RENDER = 25
    SEPARATOR = 26
    GEOMETRY_DATA = 27
    KEY_VALUE_DICT = 28
    LABEL = 29
    RGBAMASK = 30
    ORDINAL = 31
    RAMP_FLT = 32
    RAMP_RGB = 33
    FLOAT_LOG = 34
    INT_LOG = 35
    DATA = 36
    FLOAT_MINMAX = 37
    INT_MINMAX = 38
    INT_STARTEND = 39
    BUTTONSTRIP = 40
    ICONSTRIP = 41
    GROUPRADIO = 1000
    GROUPCOLLAPSIBLE = 1001
    GROUPSIMPLE = 1002
    GROUP = 1003


class ChoiceListType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_ChoiceListType
    """
    NONE = 0
    NORMAL = 1
    MINI = 2
    REPLACE = 3
    TOGGLE = 4


class PresetType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_PresetType
    """
    INVALID = -1
    BINARY = 0
    IDX = 1
    MAX = 2


class NodeType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_NodeType
    """
    ANY = -1
    NONE = 0
    OBJ = 1 << 0
    SOP = 1 << 1
    CHOP = 1 << 2
    ROP = 1 << 3
    SHOP = 1 << 4
    COP = 1 << 5
    VOP = 1 << 6
    DOP = 1 << 7
    TOP = 1 << 8


class NodeFlags(enum.IntEnum):
    """Equivalent of HAPI's HAPI_NodeFlags
    """
    ANY = -1
    NONE = 0
    DISPLAY = 1 << 0
    RENDER = 1 << 1
    TEMPLATED = 1 << 2
    LOCKED = 1 << 3
    EDITABLE = 1 << 4
    BYPASS = 1 << 5
    NETWORK = 1 << 6

    OBJ_GEOMETRY = 1 << 7
    OBJ_CAMERA = 1 << 8
    OBJ_LIGHT = 1 << 9
    OBJ_SUBNET = 1 << 10

    SOP_CURVE = 1 << 11
    SOP_GUIDE = 1 << 12

    TOP_NONSCHEDULER = 1 << 13


class GroupType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_GroupType
    """
    INVALID = -1
    POINT = 0
    PRIM = 1
    MAX = 2


class AttributeOwner(enum.IntEnum):
    """Equivalent of HAPI's HAPI_AttributeOwner
    """
    INVALID = -1
    VERTEX = 0
    POINT = 1
    PRIM = 2
    DETAIL = 3
    MAX = 4

class CurveType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_CurveType
    """
    INVALID = -1
    LINEAR = 0
    NURBS = 1
    BEZIER = 2
    MAX = 3


class VolumeType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_VolumeType
    """
    INVALID = -1
    HOUDINI = 0
    VDB = 1
    MAX = 2


class StorageType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_StorageType
    """
    INVALID = -1
    INT = 0
    INT64 = 1
    FLOAT = 2
    FLOAT64 = 3
    STRING = 4
    MAX = 5


class AttributeTypeInfo(enum.IntEnum):
    """Equivalent of HAPI's HAPI_AttributeTypeInfo
    """
    INVALID = -1
    NONE = 0
    POINT = 1
    HPOINT = 2
    VECTOR = 3
    NORMAL = 4
    COLOR = 5
    QUATERNION = 6
    MATRIX3 = 7
    MATRIX = 8
    ST = 9
    HIDDEN = 10
    BOX2 = 11
    BOX = 12
    TEXTURE = 13
    MAX = 14


class HGeoType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_GeoType
    """
    INVALID = -1
    DEFAULT = 0
    INTERMEDIATE = 1
    INPUT = 2
    CURVE = 3
    MAX = 4


class PartType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_PartType
    """
    INVALID = -1
    MESH = 0
    CURVE = 1
    VOLUME = 2
    INSTANCER = 3
    BOX = 4
    SPHERE = 3
    MAX = 4


class InputType(enum.IntEnum):
    """Equivalent of HAPI's HAPI_InputType
    """
    INVALID = -1
    TRANSFORM = 0
    GEOMETRY = 1
    MAX = 2

class RSTOrder(enum.IntEnum):
    """Equivalent of HAPI's HAPI_RSTOrder
    """
    HAPI_TRS = 0
    HAPI_TSR = 1
    HAPI_RTS = 2
    HAPI_RST = 3
    HAPI_STR = 4
    HAPI_SRT = 5
    HAPI_RSTORDER_DEFAULT = 5


NP_TYPE_TO_HSTORAGE_TYPE = {
    np.dtype('int32'): StorageType.INT,
    np.dtype('int64'): StorageType.INT64,
    np.dtype('float32'): StorageType.FLOAT,
    np.dtype('float64'): StorageType.FLOAT64,
    np.dtype('bytes_'): StorageType.STRING}


class PartInfo(StructureWithEnums):
    """Equivalent of HAPI's HAPI_PartInfo
    """
    _fields_ = [('id', c_int32),
                ('nameSH', c_int32),
                ('type', c_int32),
                ('faceCount', c_int32),
                ('vertexCount', c_int32),
                ('pointCount', c_int32),
                ('attributeCounts', c_int32 * 4),
                ('isInstanced', c_bool),
                ('instancedPartCount', c_int32),
                ('instanceCount', c_int32),
                ('hasChanged', c_bool)]
    _map = {
        "type":  PartType
    }

    def __init__(self):
        """Init
        """
        super(PartInfo, self).__init__()
        self.attributeCounts[0] = 0
        self.attributeCounts[1] = 0
        self.attributeCounts[2] = 0
        self.attributeCounts[3] = 0

    @property
    def point_attrib_count(self):
        """Get point attribute count

        Returns:
            int: point attribute count
        """
        return self.attributeCounts[AttributeOwner.POINT]

    @point_attrib_count.setter
    def point_attrib_count(self, value):
        """Set point attribute count

        Args:
            value (int): point attribute count
        """
        self.attributeCounts[AttributeOwner.POINT] = value

    @property
    def vertex_attrib_count(self):
        """Get vertex attribute count

        Returns:
            int: vertex attribute count
        """
        return self.attributeCounts[AttributeOwner.VERTEX]

    @vertex_attrib_count.setter
    def vertex_attrib_count(self, value):
        """Set vertex attribute count

        Args:
            value (int): vertex attribute count
        """
        self.attributeCounts[AttributeOwner.VERTEX] = value

    @property
    def prim_attrib_count(self):
        """Get prim attribute count

        Returns:
            int: prim attribute count
        """
        return self.attributeCounts[AttributeOwner.PRIM]

    @prim_attrib_count.setter
    def prim_attrib_count(self, value):
        """Set prim attribute count

        Args:
            value (int): prim attribute count
        """
        self.attributeCounts[AttributeOwner.PRIM] = value

    @property
    def detail_attrib_count(self):
        """Get detail attribute count

        Returns:
            int: detail attribute count
        """
        return self.attributeCounts[AttributeOwner.DETAIL]

    @detail_attrib_count.setter
    def detail_attrib_count(self, value):
        """Set detail attribute count

        Args:
            value (int): detail attribute count
        """
        self.attributeCounts[AttributeOwner.DETAIL] = value


class CurveInfo(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_CurveInfo
    """
    _fields_ = [('curveType', c_int32),
                ('curveCount', c_int32),
                ('vertexCount', c_int32),
                ('knotCount', c_int32),
                ('isPeriodic', c_bool),
                ('isRational', c_bool),
                ('order', c_int32),
                ('hasKnots', c_bool)]
    _map = {
        "curveType":  CurveType
    }

class Transform(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_Transform
    """
    _fields_ = [('position', c_float * 3),
                ('rotationQuaternion', c_float * 4),
                ('scale', c_float * 3),
                ('shear', c_float * 3),
                ('rstorder', c_int32)]
    _map = {
        "rstorder":  RSTOrder
    }

    def __init__(self):
        super(Transform, self).__init__()
        self.position[0] = 0
        self.position[1] = 0
        self.position[2] = 0
        self.rotationQuaternion[0] = 0
        self.rotationQuaternion[1] = 0
        self.rotationQuaternion[2] = 0
        self.rotationQuaternion[3] = 1
        self.scale[0] = 1
        self.scale[1] = 1
        self.scale[2] = 1
        self.shear[0] = 0
        self.shear[1] = 0
        self.shear[2] = 0
        self.rstorder = RSTOrder.HAPI_RSTORDER_DEFAULT

class VolumeTileInfo(Structure): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_VolumeTileInfo
    """
    _fields_ = [('minX', c_int32),
                ('minY', c_int32),
                ('minZ', c_int32),
                ('isValid', c_bool)]

class VolumeInfo(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_VolumeInfo
    """
    _fields_ = [('nameSH', c_int32),
                ('type', c_int32),
                ('xLength', c_int32),
                ('yLength', c_int32),
                ('zLength', c_int32),
                ('minX', c_int32),
                ('minY', c_int32),
                ('minZ', c_int32),
                ('tupleSize', c_int32),
                ('storage', c_int32),
                ('tileSize', c_int32),
                ('transform', Transform),
                ('hasTaper', c_bool),
                ('xTaper', c_float),
                ('yTaper', c_float)]
    _map = {
        "type":  VolumeType,
        "storage": StorageType,
    }

class AttributeInfo(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_AttributeInfo
    """
    _fields_ = [('exists', c_bool),
                ('owner', c_int32),
                ('storage', c_int32),
                ('originalOwner', c_int32),
                ('count', c_int32),
                ('tupleSize', c_int32),
                ('typeInfo', c_int32)]
    _map = {
        "owner": AttributeOwner,
        "storage": StorageType,
        "originalOwner": AttributeOwner,
        "typeInfo": AttributeTypeInfo
    }


class SessionConnectionState(enum.IntEnum):
    """[summary]
    """
    NOT_CONNECTED = 0
    CONNECTED = 1
    FAILED_TO_CONNECT = 2


class ThriftServerOptions(Structure): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_ThriftServerOptions
    """
    _fields_ = [('autoClose', c_bool),
                ('timeoutMs', c_float)]


class Session(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_Session
    """
    _fields_ = [('type', c_int),
                ('id', c_int64)]
    _map = {
        "type":  SessionType
    }


class AssetInfo(Structure): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_AssetInfo
    """
    _fields_ = [('nodeId', c_int32),
                ('objectNodeId', c_int32),
                ('hasEverCooked', c_bool),
                ('nameSH', c_int32),
                ('labelSH', c_int32),
                ('filePathSH', c_int32),
                ('versionSH', c_int32),
                ('fullOpNameSH', c_int32),
                ('helpTextSH', c_int32),
                ('helpURLSH', c_int32),
                ('objectCount', c_int32),
                ('handleCount', c_int32),
                ('transformInputCount', c_int32),
                ('geoInputCount', c_int32),
                ('geoOutputCount', c_int32),
                ('haveObjectsChanged', c_int32),
                ('haveMaterialsChanged', c_int32)]


class ObjectInfo(Structure): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_ObjectInfo
    """
    _fields_ = [('nameSH', c_int32),
                ('objectInstancePathSH', c_int32),
                ('hasTransformChanged', c_bool),
                ('haveGeosChanged', c_bool),
                ('isVisible', c_bool),
                ('isInstancer', c_bool),
                ('isInstanced', c_bool),
                ('geoCount', c_int32),
                ('nodeId', c_int32),
                ('objectToInstanceId', c_int32)]


class GeoInfo(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_GeoInfo
    """
    _fields_ = [('type', c_int32),
                ('nameSH', c_int32),
                ('nodeId', c_int32),
                ('isEditable', c_bool),
                ('isTemplated', c_bool),
                ('isDisplayGeo', c_bool),
                ('hasGeoChanged', c_bool),
                ('hasMaterialChanged', c_bool),
                ('pointGroupCount', c_int32),
                ('primitiveGroupCount', c_int32),
                ('partCount', c_int32)]
    _map = {
        "type": HGeoType
    }


class CookOptions(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_CookOptions
    """
    _fields_ = [('splitGeosByGroup', c_bool),
                ('splitGeosByAttribute', c_bool),
                ('splitAttrSH', c_int32),
                ('maxVerticesPerPrimitive', c_int32),
                ('refineCurveToLinear', c_bool),
                ('curveRefineLOD', c_float),
                ('clearErrorsAndWarnings', c_bool),
                ('cookTemplatedGeos', c_bool),
                ('splitPointsByVertexAttributes', c_bool),
                ('packedPrimInstancingMode', c_int32),
                ('handleBoxPartTypes', c_bool),
                ('handleSpherePartTypes', c_bool),
                ('checkPartChanges', c_bool),
                ('extraFlags', c_int32)]
    _map = {
        'packedPrimInstancingMode' : PackedPrimInstancingMode
    }

class NodeInfo(StructureWithEnums): # pylint: disable=too-few-public-methods
    """Equivalent of HAPI's HAPI_NodeInfo
    """
    _fields_ = [('id', c_int32),
                ('parentId', c_int32),
                ('nameSH', c_int32),
                ('type', c_int),
                ('isValid', c_bool),
                ('totalCookCount', c_int32),
                ('uniqueHoudiniNodeId', c_int32),
                ('internalNodePathSH', c_int32),
                ('parmCount', c_int32),
                ('parmIntValueCount', c_int32),
                ('parmFloatValueCount', c_int32),
                ('parmStringValueCount', c_int32),
                ('parmChoiceCount', c_int32),
                ('childNodeCount', c_int32),
                ('inputCount', c_int32),
                ('outputCount', c_int32),
                ('createdPostAssetLoad', c_bool),
                ('isTimeDependent', c_bool)]
    _map = {
        "type": NodeType
    }


class ParmInfo(StructureWithEnums):
    """Equivalent of HAPI's HAPI_ParmInfo
    """
    _fields_ = [('id', c_int32),
                ('parentId', c_int32),
                ('childIndex', c_int32),
                ('type', c_int),
                ('scriptType', c_int),
                ('typeInfoSH', c_int32),
                ('permissions', c_int32),
                ('tagCount', c_int32),
                ('size', c_int32),
                ('choiceListType', c_int32),
                ('choiceCount', c_int32),
                ('nameSH', c_int32),
                ('labelSH', c_int32),
                ('templateNameSH', c_int32),
                ('helpSH', c_int32),
                ('hasMin', c_bool),
                ('hasMax', c_bool),
                ('hasUIMin', c_bool),
                ('hasUIMax', c_bool),
                ('min', c_float),
                ('max', c_float),
                ('UIMin', c_float),
                ('UIMax', c_float),
                ('invisible', c_bool),
                ('disabled', c_bool),
                ('spare', c_bool),
                ('joinNext', c_bool),
                ('labelNone', c_bool),
                ('intValuesIndex', c_int32),
                ('floatValuesIndex', c_int32),
                ('stringValuesIndex', c_int32),
                ('choiceIndex', c_int32),
                ('inputNodeType', c_int32),
                ('inputNodeFlag', c_int32),
                ('isChildOfMultiParm', c_bool),
                ('instanceNum', c_int32),
                ('instanceLength', c_int32),
                ('instanceCount', c_int32),
                ('instanceStartOffset', c_int32),
                ('rampType', c_int32),
                ('visibilityConditionSH', c_int32),
                ('disabledConditionSH', c_int32)]
    _map = {
        "type": ParmType,
        "scriptType": PrmScriptType,
        "permissions": Permissions,
        "choiceListType": ChoiceListType,
        "inputNodeType": NodeType,
        "inputNodeFlag": NodeFlags,
        "rampType": RampType
    }


    def is_int(self):
        """If this attribute is int type

        Returns:
            bool: If this attribute is int type
        """
        return (self.type >= ParmType.INT_START and
                self.type <= ParmType.INT_END)\
            or self.type == ParmType.MULTIPARMLIST\
            or self.type == ParmType.FOLDERLIST_RADIO

    def is_float(self):
        """If this attribute is float type

        Returns:
            bool: If this attribute is float type
        """
        return self.type >= ParmType.FLOAT_START and\
            self.type <= ParmType.FLOAT_END

    def is_string(self):
        """If this attribute is string type

        Returns:
            bool: If this attribute is string type
        """
        return (self.type >= ParmType.STRING_START and
                self.type <= ParmType.STRING_END)\
            or self.type == ParmType.LABEL\
            or self.type == ParmType.PATH_FILE_DIR

    def is_path(self):
        """If this attribute is path type

        Returns:
            bool: If this attribute is path type
        """
        return (self.type >= ParmType.PATH_START and
                self.type <= ParmType.PATH_END)\
            or self.type == ParmType.PATH_FILE_DIR

    def is_node(self):
        """If this attribute is node type

        Returns:
            bool: If this attribute is node type
        """
        return self.type >= ParmType.NODE_START and\
            self.type <= ParmType.NODE_END

    def is_non_value(self):
        """If this attribute is non-value

        Returns:
            bool: If this attribute is non-value
        """
        return self.type >= ParmType.NONVALUE_START and\
            self.type <= ParmType.NONVALUE_END



class ParmChoiceInfo(StructureWithEnums):
    """Equivalent of HAPI's HAPI_ParmChoiceInfo
    """
    _fields_ = [
        ('parentParmId', c_int32),
        ('labelSH', c_int32),
        ('valueSH', c_int32)]



# -*- coding: utf-8 -*-
"""Wrapper for houdini engine's geometry.
Author  : Maajor
Email   : info@ma-yidong.com

HGeo:
    An base class for houdini engine's geometry, including shared operation\
        for setting and getting attributes. It could derived HGeoMesh\
        for handling mesh, HGeoCurve for handling curve, HGeoVolume for \
            handling volume data.

HGeoMesh:
    An object containing mesh data

HGeoCurve:
    An object containing curve data

Example usage:

import pyhapi as ph

#create houdini engine session
session = ph.HSessionManager.get_or_create_default_session()

#create an inputnode where you can set geometry
geo_inputnode = ph.HInputNode(session, "Cube")

#create a geomesh
cube_geo = ph.HGeoMesh(
    vertices=np.array(
        [[0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 1.0, 0.0],
            [0.0, 1.0, 1.0],
            [1.0, 0.0, 0.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0]], dtype=np.float32),
    faces=np.array(
        [[0, 2, 6, 4],
            [2, 3, 7, 6],
            [2, 0, 1, 3],
            [1, 5, 7, 3],
            [5, 4, 6, 7],
            [0, 4, 5, 1]], dtype=np.int32))

#set this geomesh as geometry of inputnode
geo_inputnode.set_geometry(cube_geo)

"""
import numpy as np
import logging
import functools
import operator
from ctypes import c_float

from . import hdata as HDATA
from . import hapi as HAPI

class HGeo():

    """An base class for houdini engine's geometry, including shared operation\
        for setting and getting attributes. It could derived HGeoMesh\
            for handling mesh, HGeoCurve for handling curve, HGeoVolume for \
                handling volume data.

    Attributes:
        part_info (PartInfo): geom info of this part
        point_count (int): number of points in this geo
        vertex_count (int): number of vertices in this geo
        face_count (int): number of faces in this geo
        detail_count (int): number of details in this geo, should be 1
        attribs (dict((int,str),(AttributeInfo,str,np.ndarray)): attribute's name\
            to attribute's actual data
        type_to_add_attrib (dict(int,func)): attribute's type to the function to \
            set the attribute data into hengine
    """

    def __init__(self):
        self.part_info = HDATA.PartInfo()
        self.point_count = 0
        self.vertex_count = 0
        self.face_count = 0
        self.detail_count = 1
        self.attribs = {}

        self.type_to_add_attrib = {
            HDATA.AttributeOwner.VERTEX: self.add_vertex_attrib,
            HDATA.AttributeOwner.POINT: self.add_point_attrib,
            HDATA.AttributeOwner.PRIM: self.add_prim_attrib,
            HDATA.AttributeOwner.DETAIL: self.add_detail_attrib}

    def add_attrib(self, attrib_type, name, data):
        """Add attribute data to geo, should provide attribute's\
            type, name and data

        Args:
            attrib_type (AttributeOwner): Type of the attribute
            name (str): name of the attribute
            data (ndarray(,)): Attribute data, should be 2D, 1st dims\
                corresponding to each item in that attribute, ptnum/vtnum\
                    primnum etc, 2nd dim should be tuple size of this attribute
        """
        self.type_to_add_attrib[attrib_type](name, data)

    def add_point_attrib(self, name, data):
        """Add point attribute data to geo

        Args:
            name (str): name of the attribute
            data (ndarray(,)): Attribute data, should be 2D, 1st dims\
                equals number of points, 2nd dim should be tuple size of this attribute
        """
        count, tuple_size = data.shape
        if count != self.point_count:
            logging.error("AddPointAttrib Error, Data length {0} not compatible with point count {1}".\
                format(count, self.point_count))
            return
        attrib_info = HDATA.AttributeInfo()
        attrib_info.count = count
        attrib_info.tupleSize = tuple_size
        attrib_info.exists = True
        attrib_info.storage = HDATA.NP_TYPE_TO_HSTORAGE_TYPE[data.dtype]
        attrib_info.owner = HDATA.AttributeOwner.POINT
        self.part_info.point_attrib_count += 1

        #self.Attribs.append((attribInfo, name, data))
        self.attribs[(HDATA.AttributeOwner.POINT, name)] = (
            attrib_info, name, data)

    def add_vertex_attrib(self, name, data):
        """Add vertex attribute data to geo

        Args:
            name (str): name of the attribute
            data (ndarray(,)): Attribute data, should be 2D, 1st dims\
                equals number of vertices, 2nd dim should be tuple size of this attribute
        """
        count, tuple_size = data.shape
        if count != self.vertex_count:
            logging.error("AddVertexAttrib Error, Data length {0} not compatible with vertex count {1}".\
                format(count, self.vertex_count))
            return
        attrib_info = HDATA.AttributeInfo()
        attrib_info.count = count
        attrib_info.tupleSize = tuple_size
        attrib_info.exists = True
        attrib_info.storage = HDATA.NP_TYPE_TO_HSTORAGE_TYPE[data.dtype]
        attrib_info.owner = HDATA.AttributeOwner.VERTEX
        self.part_info.vertex_attrib_count += 1

        #self.Attribs.append((attribInfo, name, data))
        self.attribs[(HDATA.AttributeOwner.VERTEX, name)] = (
            attrib_info, name, data)

    def add_prim_attrib(self, name, data):
        """Add prim attribute data to geo

        Args:
            name (str): name of the attribute
            data (ndarray(,)): Attribute data, should be 2D, 1st dims\
                equals number of faces, 2nd dim should be tuple size of this attribute
        """
        count, tuple_size = data.shape
        if count != self.face_count:
            logging.error("AddPrimAttrib Error, Data length {0} not compatible with prim count {1}".\
                format(count, self.face_count))
            return
        attrib_info = HDATA.AttributeInfo()
        attrib_info.count = count
        attrib_info.tupleSize = tuple_size
        attrib_info.exists = True
        attrib_info.storage = HDATA.NP_TYPE_TO_HSTORAGE_TYPE[data.dtype]
        attrib_info.owner = HDATA.AttributeOwner.PRIM
        self.part_info.prim_attrib_count += 1

        #self.Attribs.append((attribInfo, name, data))
        self.attribs[(HDATA.AttributeOwner.PRIM, name)] = (
            attrib_info, name, data)

    def add_detail_attrib(self, name, data):
        """Add detail attribute data to geo

        Args:
            name (str): name of the attribute
            data (ndarray(,)): Attribute data, should be 2D, 1st dims\
                should be 1, 2nd dim should be tuple size of this attribute
        """
        count, tuple_size = data.shape
        if count != self.detail_count:
            logging.error("add_detail_attrib Error, Data length {0} not compatible with detail count {1}".\
                format(count, self.detail_count))
            return
        attrib_info = HDATA.AttributeInfo()
        attrib_info.count = count
        attrib_info.tupleSize = tuple_size
        attrib_info.exists = True
        attrib_info.storage = HDATA.NP_TYPE_TO_HSTORAGE_TYPE[data.dtype]
        attrib_info.owner = HDATA.AttributeOwner.DETAIL
        self.part_info.detail_attrib_count += 1

        #self.Attribs.append((attribInfo, name, data))
        self.attribs[(HDATA.AttributeOwner.DETAIL, name)] = (
            attrib_info, name, data)

    def get_attrib_data(self, attrib_type, name):
        """Get attribute data of certain type and name

        Args:
            attrib_type (AttributeOwner): Type of querying attribute
            name (str): Name of querying attribute

        Returns:
            ndarray(,): Data of querying attribute
        """
        if (attrib_type, name) in self.attribs:
            _, _, data = self.attribs[(attrib_type, name)]
            return data
        return None

    def get_attrib_tuple_size(self, attrib_type, name):
        """Get attribute data of certain type and name

        Args:
            attrib_type (AttributeOwner): Type of querying attribute
            name (str): Name of querying attribute

        Returns:
            ndarray(,): Data of querying attribute
        """
        if (attrib_type, name) in self.attribs:
            attrib_info, _, data = self.attribs[(attrib_type, name)]
            return attrib_info.tupleSize
        return None

    def get_attrib_data_storage_type(self, attrib_type, name):
        """Get attribute data of certain type and name

        Args:
            attrib_type (AttributeOwner): Type of querying attribute
            name (str): Name of querying attribute

        Returns:
            ndarray(,): Data of querying attribute
        """
        if (attrib_type, name) in self.attribs:
            attrib_info, _, data = self.attribs[(attrib_type, name)]
            return attrib_info.storage
        return None

    def get_attrib_names(self):
        """Get all attribute name in this geo

        Returns:
            list((str,AttributeOwner)): All attributes containing in this geo
        """
        attrib_names = []
        for k, _ in self.attribs.items():
            attrib_type, name = k
            attrib_names.append([name, HDATA.AttributeOwner(attrib_type)])
        return attrib_names

    def commit_to_node(self, session, node_id):
        """Set this geo into hengine's node

        Args:
            session (int64): The session of Houdini you are interacting with.
            node_id (int): The node to add geo.
        """
        HAPI.set_part_info(session.hapi_session, node_id, self.part_info)

        for attrib_info, name, data in self.attribs.values():
            HAPI.add_attribute(session.hapi_session, node_id, name, attrib_info)
            HAPI.STORAGE_TYPE_TO_SET_ATTRIB[attrib_info.storage](
                session.hapi_session, node_id, name, attrib_info, data)

    def extract_from_sop(self, session, part_info, node_id, part_id=0):
        """Extract geometry from sop

        Args:
            session (int64): The session of Houdini you are interacting with.
            part_info (PartInfo): The info of part
            node_id (int): The node to add geo.
            part_id (int): Part id. Default to 0
        """
        self.part_info = part_info
        self.point_count = part_info.pointCount
        self.vertex_count = part_info.vertexCount
        self.face_count = part_info.faceCount

        #Fill attributes
        for attrib_type in range(0, HDATA.AttributeOwner.MAX):
            attrib_names = HAPI.get_attribute_names(
                session.hapi_session,
                node_id,
                self.part_info,
                attrib_type)
            for attrib_name in attrib_names:
                # do not extract private data
                if not attrib_name.startswith("__"):
                    attrib_info = HAPI.get_attribute_info(
                        session.hapi_session, node_id, part_id, attrib_name, attrib_type)
                    data = HAPI.STORAGE_TYPE_TO_GET_ATTRIB[attrib_info.storage](
                        session.hapi_session, node_id, part_id, attrib_name, attrib_info)
                    self.attribs[(attrib_type, attrib_name)] = (
                        attrib_info, attrib_name, data)

class HGeoInstancer(HGeo):
    def __init__(self):
        super(HGeoInstancer, self).__init__()
        self.instanced_geos = []
        self.transforms = []

class HGeoMesh(HGeo):

    """A class representing hengine's mesh geometry

    Attributes:
        faces (np.ndarray): Faces data, should be in 2D such as\
                (face_count, vertex_each_face).
    """

    def __init__(self, vertices=None, faces=None):
        """Initialize

        Args:
            vertices (np.ndarray, optional): Verticed data, should be 2D:\
                (pount_count, 3). Defaults to None.
            faces (np.ndarray, optional): Faces data, should be in 2D such as\
                (face_count, vertex_each_face). Defaults to None.
        """
        super(HGeoMesh, self).__init__()
        if isinstance(vertices, np.ndarray):
            self.point_count = vertices.shape[0]

            if isinstance(faces, np.ndarray):            
                self.face_count = faces.shape[0]                
                self.faces = faces
                self.vertex_count = faces.flatten().shape[0]     
            elif type(faces) is list:
                self.face_count = len(faces)
                self.faces = faces
                self.vertices = np.array(functools.reduce(operator.concat, faces), dtype=np.int32)
                self.vertex_count = self.vertices.shape[0]
                self.face_counts = np.array([len(face) for face in self.faces], dtype=np.int32)

            self.part_info.vertexCount = self.vertex_count
            self.part_info.pointCount = self.point_count
            self.part_info.faceCount = self.face_count
            self.part_info.type = HDATA.PartType.MESH
            self.add_attrib(HDATA.AttributeOwner.POINT, "P", vertices)


    def extract_from_sop(self, session, part_info, node_id, part_id=0):
        """Extract mesh from sop

        Args:
            session (int64): The session of Houdini you are interacting with.
            part_info (PartInfo): The info of part
            node_id (int): The node to add geo.
            part_id (int): Part id. Default to 0
        """
        super().extract_from_sop(session, part_info, node_id, part_id)

        self.faces = HAPI.get_faces(session.hapi_session, node_id, part_info)

    def commit_to_node(self, session, node_id):
        """Set this geo into hengine's node

        Args:
            session (int64): The session of Houdini you are interacting with.
            node_id (int): The node to add geo.
        """
        super().commit_to_node(session, node_id)

        if type(self.faces) is list:
            HAPI.set_vertex_list(session.hapi_session, node_id, self.vertices)
            HAPI.set_face_counts(session.hapi_session, node_id, self.face_counts)
        else:
            HAPI.set_vertex_list(session.hapi_session, node_id, self.faces)
            HAPI.set_face_counts(session.hapi_session, node_id, \
                np.array([len(face) for face in self.faces]))
        HAPI.commit_geo(session.hapi_session, node_id)


class HGeoCurve(HGeo):

    """A class representing hengine's curve geometry

    Attributes:
        curve_knots (np.ndarray): Curve knots
        curve_count (np.ndarray): Curve counts
        curve_info (CurveInfo): CurveInfo
    """

    def __init__(self, vertices=None, curve_knots=None, # pylint: disable=too-many-arguments
                 is_periodic=False, order=4, curve_type=HDATA.CurveType.LINEAR):
        """Initialize

        Args:
            vertices (ndarray): Position of curve cvs, should be in 2D (vertices_count, 3)
            curve_knots (ndarray, optional): Knots of cvs. Defaults to None.
            order (int, optional): Order of curve. Defaults to 4.
            curve_type (CurveType, optional): Type of curve. \
                Defaults to HDATA.CurveType.LINEAR.
        """
        super(HGeoCurve, self).__init__()
        if isinstance(vertices, np.ndarray):
            self.point_count = vertices.shape[0]
            self.vertex_count = vertices.shape[0]
            self.face_count = 1
            self.curve_knots = curve_knots
            self.curve_count = np.repeat(vertices.shape[0], 1)

            self.part_info.type = HDATA.PartType.CURVE
            self.part_info.faceCount = self.face_count
            self.part_info.vertexCount = self.vertex_count
            self.part_info.pointCount = self.point_count

            self.curve_info = HDATA.CurveInfo()
            self.curve_info.curveType = curve_type
            self.curve_info.curveCount = 1
            self.curve_info.vertexCount = vertices.shape[0]
            self.curve_info.knotCount = (
                curve_knots.shape[0] if isinstance(curve_knots, np.ndarray) else 0)
            self.curve_info.isPeriodic = is_periodic
            self.curve_info.order = order
            self.curve_info.hasKnots = isinstance(curve_knots, np.ndarray)

            #self.AddPointAttrib("P", vertices)
            self.add_attrib(HDATA.AttributeOwner.POINT, "P", vertices)

    def commit_to_node(self, session, node_id):
        """Set this geo into hengine's node

        Args:
            session (int64): The session of Houdini you are interacting with.
            node_id (int): The node to add geo.
        """
        super().commit_to_node(session, node_id)

        HAPI.set_curve_info(session.hapi_session, node_id, self.curve_info)
        HAPI.set_curve_counts(session.hapi_session, node_id,\
            self.part_info.id, self.curve_count)
        HAPI.set_curve_knots(session.hapi_session, node_id,\
            self.part_info.id, self.curve_knots)

        HAPI.commit_geo(session.hapi_session, node_id)

    def extract_from_sop(self, session, part_info, node_id, part_id=0):
        """Extract curve from sop

        Args:
            session (int64): The session of Houdini you are interacting with.
            part_info (PartInfo): The info of part
            node_id (int): The node to add geo.
            part_id (int): Part id. Default to 0
        """
        super().extract_from_sop(session, part_info, node_id, part_id)
        self.curve_info = HAPI.get_curve_info(session.hapi_session, node_id, part_info.id)
        self.point_count = self.curve_info.vertexCount
        self.vertex_count = self.curve_info.vertexCount
        self.face_count = self.curve_info.curveCount
        self.curve_count = HAPI.get_curve_counts(session.hapi_session, node_id, \
            part_info.id, self.face_count)
        if self.curve_info.hasKnots:
            self.curve_knots = HAPI.get_curve_knots(session.hapi_session, node_id, \
                part_info.id, self.curve_info.knotCount)

class HGeoVolume(HGeo):
    """[summary]

    Args:
        HGeo ([type]): [description]
    """

    def __init__(self):
        """Initialize

        Args:
            vertices (np.ndarray, optional): Verticed data, should be 2D:\
                (pount_count, 3). Defaults to None.
            faces (np.ndarray, optional): Faces data, should be in 2D such as\
                (face_count, vertex_each_face). Defaults to None.
        """
        super(HGeoVolume, self).__init__()

    def extract_from_sop(self, session, part_info, node_id, part_id=0):
        """Extract mesh from sop

        Args:
            session (int64): The session of Houdini you are interacting with.
            part_info (PartInfo): The info of part
            node_id (int): The node to add geo.
            part_id (int): Part id. Default to 0
        """
        super().extract_from_sop(session, part_info, node_id, part_id)

    def commit_to_node(self, session, node_id):
        """Set this geo into hengine's node

        Args:
            session (int64): The session of Houdini you are interacting with.
            node_id (int): The node to add geo.
        """
        super().commit_to_node(session, node_id)

class HGeoHeightfield(HGeo):
    """[summary]

    Args:
        HGeo ([type]): [description]
    """

    def __init__(self, volume=None, volume_name="height", transform=None):
        """Initialize

        Args:
            vertices (np.ndarray, optional): Verticed data, should be 2D:\
                (pount_count, 3). Defaults to None.
            faces (np.ndarray, optional): Faces data, should be in 2D such as\
                (face_count, vertex_each_face). Defaults to None.
        """
        super(HGeoHeightfield, self).__init__()
        if isinstance(volume, np.ndarray):
            self.volume = volume
            self.volume_name = volume_name
            self.transform = transform
            self.xsize, self.ysize, self.tuple_size = volume.shape
            self.zsize = 1

            self.part_info.type = HDATA.PartType.VOLUME
            self.part_info.faceCount = 1
            self.part_info.vertexCount = 1
            self.part_info.pointCount = 1

            self.volume_info = HDATA.VolumeInfo()
            self.volume_info.xLength = self.xsize
            self.volume_info.yLength = self.ysize
            self.volume_info.zLength = self.zsize
            self.volume_info.tupleSize = self.tuple_size
            self.volume_info.tileSize = 8
            self.volume_info.type = HDATA.VolumeType.HOUDINI
            self.volume_info.storage = (
                HDATA.StorageType.FLOAT if volume.dtype is np.dtype('float32') \
                    else HDATA.StorageType.INT)
            if isinstance(transform, HDATA.Transform):
                self.volume_info.transform = transform
            else:
                tr = HDATA.Transform()
                tr.scale[0] = self.xsize
                tr.scale[1] = self.ysize
                tr.scale[2] = self.zsize/2.0
                self.volume_info.transform = tr
                self.transform = tr

    def extract_from_sop(self, session, part_info, node_id, part_id=0):
        """Extract mesh from sop

        Args:
            session (int64): The session of Houdini you are interacting with.
            part_info (PartInfo): The info of part
            node_id (int): The node to add geo.
            part_id (int): Part id. Default to 0
        """
        super().extract_from_sop(session, part_info, node_id, part_id)
        self.volume_info = HAPI.get_volume_info(session.hapi_session, node_id, part_id)
        self.xsize = self.volume_info.xLength
        self.ysize = self.volume_info.yLength
        self.zsize = self.volume_info.zLength
        self.tuple_size = self.volume_info.tupleSize
        self.volume = HAPI.get_heightfield_data(session.hapi_session, \
            node_id, part_id, self.volume_info)
        self.volume_name = HAPI.get_string(session.hapi_session, self.volume_info.nameSH)

    def commit_to_node(self, session, node_id):
        """Set this geo into hengine's node

        Args:
            session (int64): The session of Houdini you are interacting with.
            node_id (int): The node to add geo.
        """
        geo_info = HAPI.get_geo_info(session.hapi_session, node_id)
        volume_info = HAPI.get_volume_info(session.hapi_session, geo_info.nodeId, 0)

        assert volume_info.xLength == self.xsize and volume_info.yLength == self.ysize \
            and volume_info.zLength == 1 and volume_info.tupleSize == self.tuple_size
        volume_info.tileSize = 1
        volume_info.type = HDATA.VolumeType.HOUDINI
        volume_info.transform = self.transform
        self.volume_info = volume_info

        #super().commit_to_node(session, geo_info.nodeId)

        HAPI.set_volume_info(session.hapi_session, geo_info.nodeId, 0, volume_info)
        HAPI.set_heightfield_data(session.hapi_session, node_id, 0, self.volume_name, self.volume)
        HAPI.commit_geo(session.hapi_session, node_id)



# -*- coding: utf-8 -*-
"""Wrapper for houdini engine's node
Author  : Maajor
Email   : info@ma-yidong.com

HNodeBase:
    A base class for houdini engine's node, including shared operation\
        for setting input/output, setting/getting geometry, get node \
            information etc.

HNode:
    A node user created for operator name or hda assets, could get/set\
        params, press button and cook.

HInputNode:
    A node dedicated to marshall geom datas

HExistingNode:
    A node created by hengine itself, such as child node of node instantiated\
        with HNode

Example usage:

import pyhapi as ph

#create houdini engine session
session = ph.HSessionManager.get_or_create_default_session()

#create an inputnode where you can set geometry
geo_inputnode = ph.HInputNode(session, "Cube")

#create an geo sop
another_box = ph.HNode(session, "geo", "ProgrammaticBox", parent_node=asset_node)

#connect input node into geo sop's input
another_box.connect_node_input(geo_inputnode)

#disconnect input node
another_box.disconnect_node_input(0).delete()

"""
import logging
import traceback
from . import hapi as HAPI
from . import hdata as HDATA
from .hgeo import HGeoMesh, HGeoCurve, HGeo, HGeoHeightfield, HGeoVolume, HGeoInstancer

class HNodeBase():
    """A base class for houdini engine's node, including shared operation\
        for setting input/output, setting/getting geometry, get node \
            information etc.

    Attributes:
        session (HSession): HEngine session containing this node
        instantiated (bool): instantiated
        node_id (int): Current node id
        name (str): Current node name
    """

    def __init__(self, session):
        self.session = session
        self.instantiated = False
        self.node_id = -1
        self.name = ""
        self.path = ""
        self.node_info = HDATA.NodeInfo()
        self.param_info = []
        self.param_id_dict = {}
        self.param_choice_lists = {}

    def is_inited(self):
        """If this node is inited

        Returns:
            bool: If this node is inited
        """
        if not self.instantiated:
            logging.warn("Asset Not Instantiated")
        return self.instantiated

    def get_node_type(self):
        """Get node's type

        Returns:
            NodeType: type of node
        """
        return self.node_info.type

    def connect_node_input(self, node_to_connect, input_index=0, output_index=0):
        """Connect another node into this node's input

        Args:
            node_to_connect (HNodeBase): The other node to connect to this node
            input_index (int, optional): This node's input index. Defaults to 0.
            output_index (int, optional): The other node's output index. Defaults to 0.

        Returns:
            HNodeBase: Current node itself
        """
        HAPI.connect_node_input(self.session.hapi_session,\
            self.node_id, node_to_connect.node_id, input_index, output_index)
        return self

    def disconnect_node_input(self, input_index=0):
        """Disconnet this node's input

        Args:
            input_index (int, optional): This node's input index. Defaults to 0.

        Returns:
            HNodeBase: Current node itself
        """
        HAPI.disconnect_node_input(self.session.hapi_session, self.node_id, input_index)
        return self

    def get_node_input(self, input_index=0):
        """Get information of input node

        Args:
            input_index (int, optional): This node's input index. Defaults to 0.

        Returns:
            HNodeBase: Input node
        """
        input_node_id = HAPI.query_node_input(self.session.hapi_session, self.node_id, input_index)
        return self.session.get_node(input_node_id)

    def get_node_input_name(self, input_index = 0):
        return HAPI.get_node_input_name(self.session.hapi_session, self.node_id, input_index)

    def get_child_nodes(self):
        """Get children node information

        Returns:
            [HNodeBase]: All children node of this node
        """
        child_count = HAPI.compose_child_node_list(self.session.hapi_session, self.node_id)
        child_nodes = HAPI.get_composed_child_node_list(self.session.hapi_session,\
             self.node_id, child_count)
        return [self.session.get_node(node_id) for node_id in child_nodes]

    def set_geometry(self, geo):
        """Set an HGeo to this node

        Args:
            geo (HGeo): Geometry to set into this node
        """
        geo.commit_to_node(self.session, self.node_id)

    def __get_display_geo_by_node(self, node_id):
        all_geos = []
        id2geo = {}
        instancers = []
        geo_info = HAPI.get_display_geo_info(self.session.hapi_session, node_id)
        for idx in range(0, geo_info.partCount):
            part_info = HAPI.get_part_info(self.session.hapi_session, geo_info.nodeId, idx)
            add_to_result = True
            if part_info.type == HDATA.PartType.MESH:
                extract_geo = HGeoMesh()
            elif part_info.type == HDATA.PartType.CURVE:
                extract_geo = HGeoCurve()
            elif part_info.type == HDATA.PartType.VOLUME:
                volume_info = HAPI.get_volume_info(self.session.hapi_session, \
                    geo_info.nodeId, part_info.id)
                if volume_info.zLength == 1:
                    extract_geo = HGeoHeightfield()
                else:
                    extract_geo = HGeoVolume()
            elif part_info.type == HDATA.PartType.INSTANCER:
                extract_geo = HGeoInstancer()
                instancers.append(extract_geo)
            else:
                logging.critical("Type of geo extraction not implemented {0}".format(part_info.type))
                extract_geo = HGeo()
            
            extract_geo.extract_from_sop(self.session, part_info, geo_info.nodeId, part_info.id)
            id2geo[part_info.id] = extract_geo

            if add_to_result:
                all_geos.append(extract_geo)

        for instancer in instancers:
            inst_ids = HAPI.get_instanced_part_ids(self.session.hapi_session, geo_info.nodeId, instancer.part_info.id)
            xforms = HAPI.get_instancer_part_transforms(self.session.hapi_session, \
                geo_info.nodeId, instancer.part_info.id, HDATA.RSTOrder.HAPI_RSTORDER_DEFAULT)
            instancer.transforms = xforms
            instancer.instanced_geos = [id2geo[id] for id in inst_ids]

        return all_geos

    def _collect_params(self):
        self.node_info = HAPI.get_node_info(self.session.hapi_session, self.node_id)
        # update name and path
        self.name = HAPI.get_string(self.session.hapi_session, self.node_info.nameSH)
        self.path = HAPI.get_node_path(self.session.hapi_session, self.node_id)
        
        self.param_info = HAPI.get_parameters(\
            self.session.hapi_session, self.node_id, self.node_info)
        
        self.param_id_dict.clear()
        for i in range(0, self.node_info.parmCount):
            namesh = self.param_info[i].nameSH
            namestr = HAPI.get_string(self.session.hapi_session, namesh)
            self.param_id_dict[namestr] = i

        # collect choice lists
        choice_lists = HAPI.get_parm_choice_lists(self.session.hapi_session, self.node_id)
        self.param_choice_lists.clear()
        for c in choice_lists:
            if c.parentParmId not in self.param_choice_lists:
                self.param_choice_lists[c.parentParmId] = []
            self.param_choice_lists[c.parentParmId].append(c)

    def refresh_params(self):
        self._collect_params()

    def get_param_names(self):
        """Get all param in this node

        Returns:
            [str]: Name of all params
        """
        return self.param_id_dict.keys()

    def get_param_type(self, param_name):
        """Get param's type by name

        Args:
            param_name (str): Param name

        Returns:
            type: type of param
        """
        param_idx = self.param_id_dict[param_name]
        paraminfo = self.param_info[param_idx]
        if paraminfo.is_int():
            return type(int)
        if paraminfo.is_float():
            return type(float)
        if paraminfo.is_string():
            return type(str)
        return type(None)

    def get_param_choice_list(self, param_name):
        """Get param choice list

        Args:
            param_name (str): Parameter name to retrieve

        Returns:
            [int]: param choice
        """
        param_idx = self.param_id_dict[param_name]
        param_id = self.param_info[param_idx].id
        return self.param_choice_lists[param_id]


    def get_param_value(self, param_name, tupleid=0):
        """Get param value

        Args:
            param_name (str): Parameter name to retrieve

        Returns:
            int/float/str: Value of that param, depends on param type
        """
        if not self.is_inited():
            return None
        param_idx = self.param_id_dict[param_name]
        paraminfo = self.param_info[param_idx]
        if paraminfo.is_int():
            return HAPI.get_parm_int_value(self.session.hapi_session, self.node_id, param_name, tupleid)
        if paraminfo.is_float():
            return HAPI.get_parm_float_value(self.session.hapi_session, self.node_id, param_name, tupleid)
        if paraminfo.is_string():
            return HAPI.get_parm_string_value(self.session.hapi_session, self.node_id, param_name, tupleid)
        return None

    def set_param_value(self, param_name, value, tupleid=0):
        """Set parameter value

        Args:
            param_name (str): Parameter name to set
            value (int/float/str): Value to set to that param, \
                depends on param type

        Returns:
            bool: set successed
        """
        if not self.is_inited():
            return False
        param_idx = self.param_id_dict[param_name]
        paraminfo = self.param_info[param_idx]
        try:
            if paraminfo.is_int():
                HAPI.set_parm_int_value(self.session.hapi_session, \
                    self.node_id, param_name, value, tupleid)
            elif paraminfo.is_float():
                HAPI.set_parm_float_value(\
                    self.session.hapi_session, self.node_id, param_name, value, tupleid)
            elif paraminfo.is_node() and isinstance(value, HNodeBase):
                HAPI.set_parm_node_value(self.session.hapi_session, \
                    self.node_id, param_name, value.node_id)
            elif paraminfo.is_string():
                HAPI.set_parm_string_value(self.session.hapi_session, \
                    self.node_id, paraminfo.id, value, tupleid)
            return True
        except AssertionError as error:
            logging.error("HAPI excecution failed")
            logging.error(error)
            return False

    def cook(self, cook_option : HDATA.CookOptions = None, status_report_interval=1.0, status_verbosity=HDATA.StatusVerbosity.ALL):
        """Cook this node in sync/blocking manner

        Returns:
            HNodeBase: Current node itself
        """
        if not self.is_inited():
            return None
        if cook_option == None:
            cook_option = self.session.cook_option
        HAPI.cook_node(self.session.hapi_session, cook_option, self.node_id,\
            status_report_interval, status_verbosity)
        return self

    async def cook_async(self, cook_option : HDATA.CookOptions = None, status_report_interval=1.0, status_verbosity=HDATA.StatusVerbosity.ALL):
        """Cook this node in async/non-blocking manner

        Returns:
            HNodeBase: Current node itself
        """
        if not self.is_inited():
            return
        if cook_option == None:
            cook_option = self.session.cook_option
        await HAPI.cook_node_async(self.session.hapi_session, \
            cook_option, self.node_id, status_report_interval, status_verbosity)

    def press_button(self, param_name, status_report_interval=5.0, status_verbosity=HDATA.StatusVerbosity.ALL):
        """Press button in this node in sync/blocking manner

        Args:
            param_name (str): Button name to press
            status_report_interval (float): Time interval \
                in seconds to report status
        """
        if not self.is_inited():
            return
        #paramid = self.param_id_dict[param_name]
        #paraminfo = self.param_info[paramid]
        HAPI.set_parm_int_value(self.session.hapi_session, self.node_id, param_name, 1)
        HAPI.wait_cook(self.session.hapi_session, status_report_interval, status_verbosity)
        HAPI.set_parm_int_value(self.session.hapi_session, self.node_id, param_name, 0)

    async def press_button_async(self, param_name, status_report_interval=5.0):
        """Press button in this node in async/non-blocking manner

        Args:
            param_name (str): Button name to press
            status_report_interval (float): Time interval \
                in seconds to report status
        """
        if not self.is_inited():
            return
        #paramid = self.param_id_dict[param_name]
        #paraminfo = self.param_info[paramid]
        HAPI.set_parm_int_value(self.session.hapi_session, self.node_id, param_name, 1)
        await HAPI.wait_cook_async(self.session.hapi_session, status_report_interval)
        HAPI.set_parm_int_value(self.session.hapi_session, self.node_id, param_name, 0)

    def get_display_geos(self):
        """Get display geo of this node

        Returns:
            list of HGeo: List of geos in this node
        """
        if self.get_node_type() == HDATA.NodeType.OBJ:
            all_geos = []
            child_sop_count = HAPI.compose_object_list(self.session.hapi_session, self.node_id)
            if child_sop_count == 0:
                return self.__get_display_geo_by_node(self.node_id)
            child_object_infos = HAPI.get_composed_object_list(self.session.hapi_session,\
                self.node_id, child_sop_count)
            for objectinfo in child_object_infos:
                try:
                    sop_geo = self.__get_display_geo_by_node(objectinfo.nodeId)
                    all_geos.extend(sop_geo)
                except AssertionError as error:
                    nodename = HAPI.get_string(self.session.hapi_session, objectinfo.nameSH)
                    logging.error("Operator:{0}, cannot retrieve geo, skipped".format(nodename))
            return all_geos
        if self.get_node_type() == HDATA.NodeType.SOP:
            return self.__get_display_geo_by_node(self.node_id)
        logging.error("Operator type is {0}, cannot retrieve geo".format(self.get_node_type()))
        return None

    def rename(self, new_name):
        """rename this node
        """
        HAPI.rename_node(self.session.hapi_session, self.node_id, new_name)
        self.name = new_name

    def delete(self):
        """Delete current node
        """
        try:
            HAPI.delete_node(self.session.hapi_session, self.node_id)
            self.instantiated = False
            self.session.nodes.pop(self.node_id)
        except AssertionError as error:
            logging.error("HAPI excecution failed")
            logging.error(error)


class HNode(HNodeBase):

    """A node user created for operator name or hda assets, could get/set\
        params, press button and cook.
    """
    def __init__(self, session, operator_name, node_name, parent_node=None):
        """Init

        Args:
            session (HSession): Session where this asset is loaded
            operator_name (str): Operator name of this node
            node_name (str): Name of this node
            parent_node (HNodeBase, optional): Parent node. Defaults to None.
        """
        super(HNode, self).__init__(session)
        self.node_id = HAPI.create_node(self.session.hapi_session, operator_name, node_name,\
            parent_node_id=parent_node.node_id if parent_node is not None else -1,\
                cook_on_creation=False)
        self.instantiated = True
        self.session.nodes[self.node_id] = self
        self.name = node_name
        self._collect_params()

class HInputNode(HNodeBase):

    """A node dedicated to marshall geom datas
    """

    def __init__(self, session, node_name):
        super(HInputNode, self).__init__(session)
        self.node_id = HAPI.create_input_node(self.session.hapi_session, node_name)
        self.instantiated = True
        self.name = node_name
        self.session.nodes[self.node_id] = self
        self.node_info = HAPI.get_node_info(self.session.hapi_session, self.node_id)

class HExistingNode(HNodeBase):
    """A node created by hengine itself, such as child node of node instantiated\
        with HNode
    """

    def __init__(self, session, node_id):
        super(HExistingNode, self).__init__(session)
        self.node_id = node_id
        try:
            self.node_info = HAPI.get_node_info(self.session.hapi_session, self.node_id)
            self.name = HAPI.get_string(self.session.hapi_session, self.node_info.nameSH)
            self._collect_params()
            self.instantiated = True
            self.session.nodes[self.node_id] = self
        except AssertionError as error:
            logging.error("HAPI excecution failed")
            self.instantiated = False
            traceback.print_tb(error.__traceback__)
            logging.error(error)

class HHeightfieldInputNode(HNodeBase):
    """A node dedicated to marshall heightfield datas
    """

    def __init__(self, session, node_name, x_size, y_size, voxel_size):# pylint: disable=too-many-arguments
        super(HHeightfieldInputNode, self).__init__(session)
        self.node_id, self.height_id, self.mask_id, self.merge_id\
            = HAPI.create_heightfield_input_node(self.session.hapi_session, node_name,\
            x_size, y_size, voxel_size)
        self.height_node = HExistingNode(session, self.height_id)
        self.mask_node = HExistingNode(session, self.mask_id)
        self.merge_node = HExistingNode(session, self.merge_id)
        self.instantiated = True
        self.name = node_name
        self.session.nodes[self.node_id] = self
        self.node_info = HAPI.get_node_info(self.session.hapi_session, self.node_id)
        self._collect_params()
        #hacking.....otherwise voxel size will be 0
        self.set_param_value("gridspacing", voxel_size)
        self.cook()
        self.height_node.cook()
        self.mask_node.cook()

class HHeightfieldInputVolumeNode(HNodeBase):
    """A node dedicated to marshall heightfield volume datas

    """

    def __init__(self, session, node_name, x_size, y_size, voxel_size):# pylint: disable=too-many-arguments
        super(HHeightfieldInputVolumeNode, self).__init__(session)
        self.node_id = HAPI.create_heightfield_volume_input_node(self.session.hapi_session, \
            node_name, x_size, y_size, voxel_size)
        self.instantiated = True
        self.name = node_name
        self.session.nodes[self.node_id] = self
        self.node_info = HAPI.get_node_info(self.session.hapi_session, self.node_id)
        self._collect_params()
        #hacking.....otherwise voxel size will be 0
        self.set_param_value("divsize", voxel_size)
        self.cook()



# -*- coding: utf-8 -*-
"""Interface for interacting with houdini digital assets (hda)
Author  : Maajor
Email   : info@ma-yidong.com

HAsset:
    Representing an HDA asset

Example usage:

import pyhapi as ph

#create a houdini engine session
session = ph.HSessionManager.get_or_create_default_session()

hda_asset = ph.HAsset(session, "hda/FourShapes.hda")
asset_node = hda_asset.instantiate(node_name="Processor").cook()
"""
from . import hapi as HAPI
from .hnode import HNode
from ctypes import c_int32

class HAsset():
    """An class representing an HDA asset.

    Attributes:
        instantiated (bool): if this asset has instantiated node
        hda_path (str): Path of this asset
        session (HSession): Session where this asset is loaded
        asset_names ([str]): names of operator in this asset
    """

    def __init__(self, session, hdapath):
        """Initialize

        Args:
            session (int64): The session of Houdini you are interacting with.
            hdapath (str): Path of loading hda
        """
        self.instantiated = False
        self.hda_path = hdapath
        self.session = session
        self.library_id = 0
        session.reload_asset_library(self)


    def instantiate(self, node_name="Node", operator_id=0):
        """Instantiate an operator in this node

        Args:
            node_name (str, optional): Assign a name for this node. Defaults to "Node".
            operator_id (int, optional): Operator id you want to instantiate in this asset. \
                Defaults to 0.

        Returns:
            HNode: Node instantiated
        """
        node = HNode(self.session, self.asset_names[operator_id], node_name)
        return node

    @property
    def asset_names(self):
        return HAPI.get_available_assets(
            self.session.hapi_session, self.library_id)

    def reload(self):
        self.session.reload_asset_library(self)

    def get_assets_names(self):
        """Get all operator names in this asset
        """
        return self.asset_names



# -*- coding: utf-8 -*-
"""wrapper for houdini engine's session.
Author  : Maajor
Email   : info@ma-yidong.com

HSession:
    A wrapper fou houdini engine's session, it contains the session itself\
         and other informations such as nodes, process id etc.
    Use this object to get nodes, save hips. Creating nodes and\
         lots of hengine operation need HSession object as parameter.

HSessionManager:
    Use this object to get HSession object

Example usage:

import pyhapi as ph

#create a houdini engine session
session = ph.HSessionManager.get_or_create_default_session()

#create a geo objnode
box = ph.HNode(session, "geo", "ProgrammaticBox")

#save the session to a hip file
session.save_hip("test.hip")

"""
import os
import logging
import asyncio
import threading
import enum
import time
import pyhapi
from . import hdata as HDATA
from . import hapi as HAPI
from .hnode import HExistingNode


__all__ = [
    # Classes
    'HSession', 'HSessionManager', 'HSessionPool', 'HSessionTask'
]


class HSession():

    """A wrapper fou houdini engine's session, it contains the session itself\
        and other informations such as nodes, process id etc.

    Attributes:
        hapi_session (HAPI_SessionId:int64): hapi session id
        connected_state (hdata.SessionConnectionState): state of current session
        nodes (list(hnode.HNodeBase)): all nodes in this session
        process_id (int): process id used by this session
        cook_option (hdata.CookOptions): cook option of this session
        root_path (str): file path to hsession project's root path, \
                it could contain /hda folder
    """

    def __init__(self):
        """Initialize the session
        """
        self.hapi_session = HDATA.Session(HDATA.SessionType.THRIFT, 0)
        self.connected_state = HDATA.SessionConnectionState.NOT_CONNECTED
        self.nodes = {}
        self.process_id = -1
        self.cook_option = HAPI.get_cook_options()
        self.root_path = ""
        self.pipe_name = "hapi"

        self.asset_libs = {}

    def get_node(self, node_id):
        """Get node in this session by HAPI_NodeId

        Args:
            node_id (hdata.HAPI_NodeId ): node's id store by the session

        Returns:
            hnode.HNodeBase: an object represent that node in session. It will\
                return None if cannot find any node matching the node id
        """
        try_get_node = self.nodes.get(node_id)
        if try_get_node is not None:
            return try_get_node

        existing_node = HExistingNode(self, node_id)
        if existing_node is not None:
            self.nodes[node_id] = existing_node
            return existing_node
        return None

    def reload_asset_library(self, asset):
        asset.session = self
        asset_lib_id = HAPI.load_asset_library_from_file(
            self.hapi_session, asset.hda_path)

        asset.library_id = asset_lib_id
        self.asset_libs[asset.library_id.value] = asset

        # asset.asset_names = HAPI.get_available_assets(
        #    self.hapi_session, asset_lib_id)


    def create_thrift_pipe_session(self, rootpath, pipe_name, auto_close=True, timeout=10000.0):
        """Create the session in thrift-pipe manner

        Args:
            rootpath (string): file path to hsession project's root path, \
                it could contain /hda folder
            auto_close (bool, optional): Close the server automatically when\
                 all clients disconnect from it. Defaults to True.
            timeout (float, optional): Timeout in milliseconds for waiting on \
                the server to signal that it's ready to serve. If the server \
                    fails to signal within this time interval, the start server \
                        call fails and the server process is terminated. \
                            Defaults to 10000.0.

        Returns:
            bool: true if the session created successfully, false not.
        """
        self.root_path = rootpath
        self.pipe_name = pipe_name
        return self.__internal_create_thrift_pipe_session(rootpath, pipe_name, True, auto_close, timeout)

    def __internal_create_thrift_pipe_session(\
        self, rootpath, pipe_name, create_session=True, auto_close=True, timeout=10000.0):

        try:
            self.check_and_close_existing_session()

            self.hapi_session = HDATA.Session(HDATA.SessionType.THRIFT, 0)
            self.connected_state = HDATA.SessionConnectionState.FAILED_TO_CONNECT

            if create_session:
                server_options = HDATA.ThriftServerOptions(auto_close, timeout)
                self.process_id = HAPI.start_thrift_named_pipe_server(server_options, pipe_name)

            HAPI.create_thrift_named_pipe_session(self.hapi_session, pipe_name)
            self.__initialize_session(rootpath)
            return True
        except AssertionError as error:
            logging.error("HAPI excecution failed")
            logging.error(error)
            return False

    def __initialize_session(self, rootpath):
        HAPI.initialize(self.hapi_session, self.cook_option,\
            otl_search_path="{0}\\hda\\".format(rootpath))
        self.root_path = rootpath
        self.connected_state = HDATA.SessionConnectionState.CONNECTED

    def cleanup(self):
        """Clean up current session
        """
        if self.is_session_valid():
            HAPI.cleanup(self.hapi_session)

    def check_and_close_existing_session(self):
        """Close current session

        Returns:
            bool: if the close success
        """
        if self.hapi_session is not None and self.is_session_valid():
            logging.info("Close Session")
            HAPI.cleanup(self.hapi_session)
            HAPI.close_session(self.hapi_session)
        return True

    def is_session_valid(self):
        """Check if current session is valid

        Returns:
            bool: if current session is valid
        """
        if self.connected_state != HDATA.SessionConnectionState.FAILED_TO_CONNECT:
            return HAPI.is_session_valid(self.hapi_session)
        return False

    def save_hip(self, filename="debug.hip", lock_nodes=True):
        """Save current session to a hip file

        Args:
            filename (str, optional): name of saved hip file
            lock_nodes (bool, optional): Specify whether to lock all SOP nodes \
            before saving the scene file.
        """
        HAPI.save_hip_file(self.hapi_session, filename, lock_nodes)
        logging.info("Session saved to {0}".format(filename))

    def restart_session(self):
        """Restart current session

        Returns:
            HSession: session itself
        """
        HAPI.cleanup(self.hapi_session)
        self.__initialize_session(self.root_path)
        return self

    def load_hip(self, filename, cook_on_load=False):
        """Loads a .hip file into the main Houdini scene

        Args:
            filename ([type]): HIP file absolute path to load
            cook_on_load (bool, optional): \
                Set to true if you wish the nodes to cook as soon as they are created.\
                    Defaults to False.

        Returns:
            HSession: session itself
        """
        HAPI.load_hip_file(self.hapi_session, filename, cook_on_load)
        return self

    def __del__(self):
        self.check_and_close_existing_session()

class HSessionManager():

    """Use this object to get HSession object
    """

    _defaultSession = None
    _defaultSessionPool = None
    _rootpath = os.getcwd()
    _pipe_name = "hapi"

    @staticmethod
    def get_or_create_default_session(rootpath=os.getcwd(), pipe_name = "hapi"):
        """Get or create an session

        Args:
            rootpath (str, optional): file path to hsession project's root path, \
                it could contain /hda folder

        Returns:
            HSession: session created
        """
        # check if can find libHAPIL otherwise return
        assert pyhapi.__library_initialized__,  "libHAPIL not found, Please refer to https://pyhapi.readthedocs.io/en/latest/install.html to setup Houdini Engine's PATH"
        if rootpath:
            HSessionManager._rootpath = rootpath
        if pipe_name:
            HSessionManager._pipe_name = pipe_name
        if HSessionManager._defaultSession is not None and\
            HSessionManager._defaultSession.is_session_valid():
            return HSessionManager._defaultSession
        if HSessionManager._defaultSession is None or\
            HSessionManager._defaultSession.ConnectedState ==\
                HDATA.SessionConnectionState.NOT_CONNECTED:
            if HSessionManager.__create_thrift_pipe_session( \
                HSessionManager._rootpath, pipe_name=HSessionManager._pipe_name):
                return HSessionManager._defaultSession
        HSessionManager._defaultSession = None
        return None

    @staticmethod
    def __check_and_close_existing_session():
        if HSessionManager._defaultSession is not None:
            HSessionManager._defaultSession.CloseSession()
            HSessionManager._defaultSession = None

    @staticmethod
    def __create_thrift_pipe_session(rootpath, pipe_name, auto_close=True, timeout=10000.0):
        HSessionManager.__check_and_close_existing_session()
        HSessionManager._defaultSession = HSession()
        return HSessionManager._defaultSession.\
            create_thrift_pipe_session(rootpath, pipe_name, auto_close, timeout)

    @staticmethod
    def restart_session():
        """Restart default session

        Returns:
            HSession: session created
        """
        if HSessionManager._defaultSession is not None:
            return HSessionManager._defaultSession.restart_session()
        session = HSession()
        if session.create_thrift_pipe_session(HSessionManager._rootpath, HSessionManager._pipe_name, True):
            HSessionManager._defaultSession = session
        return None

    @staticmethod
    def get_or_create_session_pool(session_count=10, rootpath=os.getcwd(), pipe_name = "hapi"):
        """Restart default session

        Returns:
            HSession: session created
        """
        if HSessionManager._defaultSessionPool is not None:
            if session_count == HSessionManager._defaultSessionPool._session_count:
                return HSessionManager._defaultSessionPool.restart_session_pool()
            else:
                return HSessionManager._defaultSessionPool.resize(session_count)
        if rootpath:
            HSessionManager._rootpath = rootpath
        if pipe_name:
            HSessionManager._pipe_name = pipe_name
        session_pool = HSessionPool(session_count)
        if session_pool.create_thrift_pipe_session(HSessionManager._rootpath, HSessionManager._pipe_name, True):
            HSessionManager._defaultSessionPool = session_pool
            return HSessionManager._defaultSessionPool
        return None


class HSessionPool():

    def __init__(self, session_count, max_task_time: int = 10000):
        """Initialize the session pool
        """
        self.sessions = []
        self._session_count = session_count
        self.current_id = 0
        self.task_queue = asyncio.Queue()
        self._loop = asyncio.get_event_loop()
        self._max_task_time = max_task_time
        self._workers = None
        self._pipe_name_prefix = "hapi"
        self._root_path = os.getcwd()
        self._timeout = 10000
        self._auto_close = True

    def __iter__(self):
        return iter(self.sessions)

    def __next__(self):
        if self.current_id <= self._session_count:
            x = self.sessions[self.current_id]
            self.current_id += 1
            return x
        else:
            raise StopIteration

    def __getitem__(self, key):
        return self.sessions[key]

    # for producer to add task
    def enqueue_task(self, task, *args):
        logging.debug("enqueue task {0} with param {1}".format(task, args))
        self.task_queue.put_nowait((task, *args))

    # for producer to add task
    async def enqueue_task_async(self, task, *args):
        logging.debug("enqueue task {0} with param {1}".format(task, args))
        await self.task_queue.put((task, *args))

    def run_on_task_producer(self, producer):
        self._loop.run_until_complete(self.run_on_task_producer_async(producer))
        self._loop.close() 

    async def run_on_task_producer_async(self, producer):
        task_producer = asyncio.create_task(producer(self))
        self._workers = [asyncio.create_task(self.__worker_loop(i)) for i in range(self._session_count)]
        await asyncio.gather(*self._workers, task_producer, return_exceptions=True)

    # run all enqueued tasks by now
    def run_all_tasks(self):
        self._loop.run_until_complete(self.__run_tasks_available())
        self._loop.close() 

    def create_thrift_pipe_session(self, rootpath, pipe_name_prefix, auto_close=True, timeout=10000.0):
        valid_session = 0
        self._root_path = rootpath
        self._timeout = timeout
        self._auto_close = auto_close
        self._pipe_name_prefix = pipe_name_prefix
        for i in range(self._session_count):
            session = HSession()
            if session.create_thrift_pipe_session(rootpath, pipe_name_prefix+str(i), auto_close, timeout):
                valid_session += 1
                self.sessions.append(session)
        self._session_count = valid_session
        if valid_session>1:
            return True
        return False

    def restart_session_pool(self):
        for session in self.sessions:
            session.restart_session()

    def resize(self, session_count):
        if session_count < self._session_count:
            for i in range(session_count, self._session_count):
                self.sessions[i].check_and_close_existing_session()
            for i in range(session_count, self._session_count):
                self.sessions.remove(session_count)
            self._session_count = session_count
        elif session_count > self._session_count:
            valid_session = 0
            for i in range(self._session_count, session_count):
                session = HSession()
                if session.create_thrift_pipe_session(self._root_path, self._pipe_name_prefix+str(i), self._auto_close, self._timeout):
                    valid_session += 1
                    self.sessions.append(session)
            self._session_count += valid_session

    async def __run_tasks_available(self):
        self._workers = [asyncio.create_task(self.__worker_loop(i)) for i in range(self._session_count)]
        await self.task_queue.join()
        for worker in self._workers:
            worker.cancel()

    # https://github.com/CaliDog/asyncpool
    # one consumer worker per houdini engine session
    async def __worker_loop(self, i):
        while True:
            got_obj = False

            try:
                avail_session = self.sessions[i]
                task_to_proceed, *args = await self.task_queue.get()
                got_obj = True

                running_coro = asyncio.wait_for(task_to_proceed(avail_session, *args), self._max_task_time, loop=self._loop)
                await running_coro
            except asyncio.CancelledError:
                logging.info("Worker {0} is Cancelled".format(i))
                break
            except KeyboardInterrupt as e:
                logging.info("Worker {0} is Interrupt".format(i))
                break
            except (MemoryError, SystemExit) as e:
                logging.exception(e)
                raise
            except BaseException as e:
                logging.exception(e)
                logging.exception("Worker Call Failed")
            finally:
                if got_obj:
                    self.task_queue.task_done()

def HSessionTask(task):
    async def wrapper(*args, **kwargs):
        assert isinstance(args[0], HSession), "{0}'s first parameter should be HSession".format(task)
        await task(*args, **kwargs)
        args[0].restart_session()
    return wrapper


# -*- coding: utf-8 -*-\
"""pyhapi
Author  : Maajor
Email   : info@ma-yidong.com
"""
import logging
from .hdata import *
from .hgeo import HGeo, HGeoMesh, HGeoCurve, HGeoHeightfield, HGeoInstancer
from .hsession import HSession, HSessionManager, HSessionPool, HSessionTask
from .hnode import HNode, HInputNode, HHeightfieldInputNode, HHeightfieldInputVolumeNode
from .hasset import HAsset
from . import hapi as HAPI

__version__ = "0.0.3b0"


import platform
import os
from ctypes import cdll

def __check_libpath(libpath):
    if not (os.path.exists(libpath) and os.path.isdir(libpath)):
        return False
    
    sys_platform = platform.system()
    libdll = None
    if sys_platform == "Windows":
        libdll = "libHAPIL.dll"
    elif sys_platform == "Linux":
        libdll = "libHAPIL.so"
    else:
        libdll = "libHAPIL.dylib"

    return os.path.exists(os.path.join(libpath,libdll) )

def __ensure_hapi_path(libpath):
    path_env = os.environ['PATH'].split(';')
    if libpath or __check_libpath(libpath):
        print(not libpath)
        if not os.path.normpath(libpath) in [os.path.normpath(p) for p in path_env]:
            path_env.append(libpath)
            os.environ['PATH']=';'.join(path_env)
        return True
    else:
        return any([__check_libpath(p) for p in path_env])

# ensure "import pyhapi" will not give error even cannot find libHAPIL
# call HSessionManager.get_or_create_default_session() to initialize, make it free of calling Initialize function
__library_initialized__ = False
if __ensure_hapi_path(""):
    from . import hapi
    SYS = platform.system()
    if SYS == "Windows":
        hapi.HAPI_LIB = cdll.LoadLibrary("libHAPIL")
    elif SYS == "Linux":
        hapi.HAPI_LIB = cdll.LoadLibrary("libHAPIL.so")
    elif SYS == "Darwin":
        hapi.HAPI_LIB = cdll.LoadLibrary("libHAPIL.dylib")

    __library_initialized__ = True
    print("HAPI Found")
else:
    print("HAPI Not Found")
